import os
from os.path import expanduser

class PanExporter:
    def __init__(self, pandataset, filelocation=None):
        self.module_dir = os.path.dirname(os.path.dirname(__file__))
        self.pandataset=pandataset
        if filelocation == None:
            self.filelocation =os.path.join(expanduser("~"),'pangaeapy_export')
            try:
                os.makedirs(self.filelocation)
            except FileExistsError:
                pass
        else:
            self.filelocation = filelocation
        self.file = None
        self.logging = self.pandataset.logging
        #print(self.logging)

    #check if export is possible
    def verify(self):
        return True

    #create the export file (as IO object if possible)
    def create(self):
        return True

    #save the file  at self.filelocation
    def save(self):
        return True

    #return a string representation of the export file
    def __str__(self):
        return ''


# -*- coding: utf-8 -*-
"""
Created on Wed Sep  5 14:58:00 2018

@author: Robert Huber 
"""
import json
import linecache

from netCDF4 import date2num, Dataset,stringtochar
import pandas as pd
import re
import numpy as np
import sys
import os

from pangaeapy.exporter.pan_exporter import PanExporter


class PanNetCDFExporter(PanExporter):
    style = 'sdn'
    def PrintException(self):
        exc_type, exc_obj, tb = sys.exc_info()
        f = tb.tb_frame
        lineno = tb.tb_lineno
        filename = f.f_code.co_filename
        linecache.checkcache(filename)
        line = linecache.getline(filename, lineno, f.f_globals)
        print('EXCEPTION IN ({}, LINE {} "{}"): {}'.format(filename, lineno, line.strip(), exc_obj))
    
    def setMainVariables(self):
        #self.time_units = 'hours since 2000-01-01 00:00:00.0'
        self.time_units ='days since -4713-01-01T00:00:00Z'
        try:
            self.netcdf.title=self.pandataset.title
            self.netcdf.id=self.pandataset.doi
            if self.pandataset.topotype=='time series':
                self.netcdf.featureType='timeSeries'
            elif self.pandataset.topotype=='profile series' or self.pandataset.topotype=='vertical profile':
                self.netcdf.featureType='profile'
            #nc.platform_code='DBBH'
           # nc.featureType=self.pandataset.getGeometry()
            self.netcdf.principal_investigator=self.pandataset.authors[0].fullname
            self.netcdf.date_update=self.pandataset.date+'+02:00'
        except Exception as e:
            self.logging.append({'ERROR': 'NetCDF main variables creation failed '+str(e)})
            #print('NetCDF main variables creation failed '+str(e))
                      
    def setParameterSynonyms(self, mappingfile=['mappings','pan_mappings.json']):
        self.logging.append({'INFO': 'Trying to set synonyms and standard names'})
        #print('Trying to set synonyms and standard names')
        with open(os.path.join(self.module_dir,*mappingfile), 'r') as mappingjson:
            mapping = json.load(mappingjson)
        #mapping=pd.read_csv(mappingfile,delimiter='\t',index_col='PANGAEA ID')
        for pacronym, param in self.pandataset.params.items():
            if str(param.id) in mapping:
                #print(str(param.id)+" "+mapping.get(str(param.id)).get('SDN P01 ID'))
                cf_name=mapping.get(str(param.id)).get('CF standard name')
                self.pandataset.params[pacronym].addSynonym('CF',cf_name,unit=mapping.get(str(param.id)).get('CF unit'))
                self.pandataset.params[pacronym].addSynonym('SD',mapping.get(str(param.id)).get('SDN P01 Name'),id=mapping.get(str(param.id)).get('SDN P01 ID'),uri=mapping.get(str(param.id)).get('SDN P01 URI'),unit=mapping.get(str(param.id)).get('SDN Unit Name'), unit_id=mapping.get(str(param.id)).get('SDN Unit URN') )
        
    def cleanParameterNames(self):
        tempparams ={}
        for pk, p in self.pandataset.params.items():
            new_pk = re.sub(r'[/\s]', '_',pk)
            new_pk = re.sub(r'[\[\]]', '', new_pk)
            new_pk = re.sub(r'(?![a-zA-Z0-9_]|{MUTF8})([^\x00-\x1F/\x7F-\xFF]|{MUTF8})', '_', new_pk)
            self.pandataset.params[new_pk] = self.pandataset.params.pop(pk)
        self.pandataset.data.columns = self.pandataset.data.columns.str.replace(r'[/\s]', '_', regex=True)
        self.pandataset.data.columns = self.pandataset.data.columns.str.replace(r'[\[\]]', '', regex=True)
        self.pandataset.data.columns = self.pandataset.data.columns.str.replace('(?![a-zA-Z0-9_]|{MUTF8})([^\x00-\x1F/\x7F-\xFF]|{MUTF8})', '_',regex=True)

    def setSDNQCVariable(self, ncvarName, dims):
        ncVar = self.netcdf.createVariable(ncvarName, 'b', dims, fill_value='57')
        ncVar.long_name = 'SeaDataNet quality flag'
        ncVar.flag_values = ', '.join(map(str, self.pandataset.quality_flag_replace.values()))
        ncVar.flag_meanings = ' '.join(self.pandataset.quality_flags.values())
        ncVar.sdn_conventions_urn = 'SDN:L20::'
        ncVar.Conventions = 'SeaDataNet measurand qualifier flags'
        return ncVar

    def renameSDNDimVars(self):
        newNames= {'Date/Time':'TIME','Depth water':'DEPTH','Latitude':'LATITUDE','Longitude':'LONGITUDE'}
        for nkey, nname in newNames.items():
            self.pandataset.rename_column(nkey,nname)

    def setSDNVariablesAndValues(self, dims):
        eventCoords = self.pandataset.data.groupby(['Event']).min()
        self.cleanParameterNames()
        var_coordinates = [c for c in ['TIME','DEPTH','LATITUDE','LONGITUDE'] if c in set(self.pandataset.params.keys())]
        crsVar = self.netcdf.createVariable('crs', 'i')
        crsVar.grid_mapping_name = "latitude_longitude"
        crsVar.epsg_code = "EPSG:4326"
        crsVar.semi_major_axis = 6378137.0
        crsVar.inverse_flattening = 298.257223563
        crsVar.assignValue(0)

        for ncvarName, p in self.pandataset.params.items():
            if ncvarName not in list(self.netcdf.dimensions.keys()) and ncvarName!='Event':
                if p.type in ['gqc','qc','numeric'] or ncvarName in ['TIME']:
                    try:
                        if not self.pandataset.data[ncvarName].isnull().all():
                            if self.netcdf.featureType == 'profile':
                                #if ncvarName in ['Latitude','Longitude','Date_Time']:
                                if ncvarName in ['LATITUDE','LONGITUDE']:
                                    ncVar = self.netcdf.createVariable(ncvarName, 'f4', ['INSTANCE'])
                                    if ncvarName == 'LATITUDE':
                                        ncQCVar = self.setSDNQCVariable('POSITION_SEADATANET_QC', ['INSTANCE'])
                                        ncQCVar[:] = [9] * eventCoords['LATITUDE'].size
                                        ncVar.units = 'degrees_north'
                                        ncVar.long_name = 'Latitude'
                                        ncVar.axis = 'Y'
                                        ncVar.sdn_uom_name = 'Degrees north'
                                        ncVar.sdn_uom_urn = 'SDN:P06::DEGN'
                                        ncVar.ancillary_variables = "POSITION_SEADATANET_QC"
                                        ncVar.grid_mapping = "crs"
                                    else:
                                        ncVar.units = 'degrees_east'
                                        ncVar.long_name = 'Longitude'
                                        ncVar.axis = 'X'
                                        ncVar.sdn_uom_name = 'Degrees east'
                                        ncVar.sdn_uom_urn = 'SDN:P06::DEGE'
                                        ncVar.ancillary_variables = "POSITION_SEADATANET_QC"
                                        ncVar.grid_mapping = "crs"
                                    ncVar[:]=eventCoords[ncvarName].values
                                elif ncvarName == 'TIME':
                                    ncVar = self.netcdf.createVariable(ncvarName, 'd', ['INSTANCE'])
                                    ncVar[:]=date2num(eventCoords['TIME'].dt.to_pydatetime(),units=self.time_units,calendar='standard')
                                    ncQCVar = self.setSDNQCVariable('TIME_SEADATANET_QC', ['INSTANCE'])
                                    ncQCVar[:] = [9] * eventCoords['TIME'].size
                                    ncVar.sdn_uom_name = 'Days'
                                    ncVar.sdn_uom_urn = 'SDN:P06::UTAA'
                                    ncVar.axis = 'T'
                                    ncVar.units = self.time_units
                                    ncVar.long_name = 'Chronological Julian Date'
                                    ncVar.ancillary_variables = "TIME_SEADATANET_QC"
                                    ncVar.calendar = 'julian'
                                else:
                                    #make sure values to fill the nc var has the same shape (dimensions) as the variable
                                    dimshape=[]
                                    for dim in dims:
                                        dimshape.append(self.netcdf.dimensions[dim].size)
                                    if ncvarName.endswith('_SEADATANET_QC'):
                                        ncVar = self.setSDNQCVariable(ncvarName, dims)
                                    else:
                                        ncVar=self.netcdf.createVariable(ncvarName,'f4',dims)
                                        ncVar.coordinates = ' '.join(var_coordinates)

                                    if ncvarName == 'DEPTH':
                                        ncVar.axis = 'Z'
                                        ncVar.positive = 'down'

                                    if ncvarName+'_SEADATANET_QC' in self.pandataset.data.columns:
                                        ncVar.ancillary_variables =ncvarName+'_SEADATANET_QC'
                                    ncValues = np.reshape(self.pandataset.data[ncvarName].values, dimshape)
                                    ncVar[:] = ncValues
                            elif self.netcdf.featureType=='timeSeries':
                                ncVar=self.netcdf.createVariable(ncvarName,'f4',dims)
                                if ncvarName=='TIME':
                                    ncVar[:]=date2num(self.pandataset.data[ncvarName].dt.to_pydatetime(),units=self.time_units,calendar='standard')
                                else:
                                    ncVar[:]=self.pandataset.data[ncvarName].values
                            else:
                                self.logging.append({'ERROR':'NetCDF Feature type not supported'})
                                break   

                            #Setting the units                                    
                            if p.synonym['CF'] != None:
                                if p.synonym['CF'].get('name'):
                                    ncVar.standard_name=p.synonym['CF'].get('name')
                                    if isinstance(p.synonym['CF']['unit'], str) ==True and not hasattr(ncVar, 'units'):
                                        p.unit=p.synonym['CF']['unit']
                            if p.synonym['SD'] != None:
                                ncVar.sdn_parameter_name=p.synonym['SD']['name']
                                ncVar.sdn_parameter_urn=p.synonym['SD']['id']
                            if not hasattr(ncVar,'long_name'):
                                ncVar.long_name=p.name
                            if not hasattr(ncVar, 'units'):
                                if p.unit!=None:
                                    ncVar.units=p.unit
                                    if p.synonym.get('SD'):
                                        if  p.synonym['SD'].get('unit'):
                                            ncVar.sdn_uom_name = p.synonym['SD']['unit']
                                            ncVar.sdn_uom_urn = p.synonym['SD']['unit_id']
                                else:
                                    ncVar.units='1'
                                    ncVar.sdn_uom_name = 'Dimensionless'
                                    ncVar.sdn_uom_urn = 'SDN:P06::UUUU'

                    except Exception as e:
                        self.logging.append({'ERROR': 'NetCDF Variable creation failed for Param: ' + ncvarName + ', ERROR: ' + str(e)})
                        self.PrintException()
                        continue

    def create(self, style='pan'):
        self.style = style
        ret = None
        if isinstance(self.pandataset.data, pd.DataFrame):
            self.cleanParameterNames()
            self.setParameterSynonyms()
            if 'Event' in self.pandataset.data.columns:                
                #TODO: Check if data set is in water: depth water or negative elevation of event.                
                if self.pandataset.topotype=='time series' or self.pandataset.topotype=='profile series' or self.pandataset.topotype=='vertical profile':
                    if style=='pan':
                        self.createPANNetCDF()
                    elif style=='sdn':
                        self.createSDNNetCDF()                        
                else:
                    self.logging.append({'ERROR': 'NetCDF Variable creation failed: Invalid Topotype (has to be profile, timeseries or series of profiles) but is: '+str(self.pandataset.topotype)})
            else:
                self.logging.append({'ERROR': 'NetCDF Variable creation failed: Event column is missing'})
            ret = self.file
        return ret

    def save(self):
        if isinstance(self.file, memoryview):
            try:
                with open(os.path.join(self.filelocation,str('netcdf_'+str(self.style)+'_'+str(self.pandataset.id)+'.nc')),'wb') as f:
                    #print(f.name)
                    f.write(self.file)
                    f.close()
                    self.logging.append({'SUCCESS': 'Saved NetCDF at: ' + str(os.path.join(self.filelocation,str('netcdf_pangaea_'+str(self.pandataset.id)+'.nc')))})
                    return True
            except Exception as e:
                self.logging.append({'ERROR': 'Could not save, NetCDF: '+str(e)})
        else:
            self.logging.append({'ERROR':'Could not save, NetCDF file is not a memoryview-object'})
            return False
        #print(type(self.file))

    def __str__(self):
        if isinstance(self.netcdf, Dataset):
            try:
                return self.netcdf.tocdl()
            except Exception as e:
                self.logging.append({'ERROR': 'Could not convert NetCDF to CDL: '+str(e)})
                return ''
                
    def createSDNNetCDF(self):
        #print('SDN2');
        self.logging.append({'INFO':'Trying to create a SeaDataNet NetCDF file'})
        self.netcdf = None
        #Determine the maximum number od data rows per event -> SDN's MAXZ, MAXT
        maxr=self.pandataset.data.groupby('Event').size().max()
        try:
            #nc = Dataset(self.filelocation+'\\nc'+str(self.pandataset.id)+'_sdn.nc','w',format='NETCDF3_CLASSIC')
            self.netcdf = Dataset(str(self.pandataset.id)+'_sdn.nc',mode = 'w', memory=1028,format='NETCDF3_CLASSIC')
            self.netcdf.Conventions='SeaDataNet_1.0 CF-1.6'
            evfr=self.pandataset.getEventsAsFrame()
            MaxStrLen=dict()
            MaxStrLen['label']=evfr['label'].str.len().max()
            if evfr['campaign'].str.len().max() >0 :               
                MaxStrLen['campaign']=evfr['campaign'].str.len().max()
            cdi_id='pan_'+str(self.pandataset.id)
            MaxStrLen['cdi_id']=len(cdi_id)
            #CampaignStrLen=evfr['campaign'].str.len().max()
            #BasisStrLen=evfr['basis'].str.len().max()
            evcnt=len(evfr['label'])
            maxtype='MAXZ'
            self.setMainVariables()
            if self.pandataset.topotype=='vertical profile' or self.pandataset.topotype=='profile series':
                maxtype='MAXZ'
            if self.pandataset.topotype=='time series':
                maxtype='MAXT'
            self.netcdf.createDimension('INSTANCE',evcnt)
            self.netcdf.createDimension(maxtype,maxr)
            for sk, StrDimLen in MaxStrLen.items():
                if 'STRING'+str(StrDimLen) not in self.netcdf.dimensions:
                    self.netcdf.createDimension('STRING'+str(StrDimLen), StrDimLen)
            #print(np.array(evfr['label']))
            events=self.netcdf.createVariable('SDN_STATION', 'S1',(u'INSTANCE',u'STRING'+str(MaxStrLen['label'])))
            events.long_name='Event label'
            if 'campaign' in MaxStrLen:
                campaign=str(MaxStrLen['campaign'])
                campaigns=self.netcdf.createVariable('SDN_CRUISE', 'S1',(u'INSTANCE',u'STRING'+campaign))
                campaigns.long_name='Campaign label'
            #Add empty rows to have the same number of rows per event
            self.pandataset.data=pd.concat([
                d.reset_index(drop=True).reindex(range(maxr))
                for n, d in self.pandataset.data.groupby('Event')
                ], ignore_index=True)
            events[:]=stringtochar(np.array(evfr['label'].tolist(),'S'+str(MaxStrLen['label'])))
            if 'campaign' in MaxStrLen:
                campaigns[:]=stringtochar(np.array(evfr['campaign'].tolist(),'S'+str(MaxStrLen['campaign'])))
            
            #mandatory bottom depth
            sdn_depth=self.netcdf.createVariable('SDN_BOT_DEPTH','f4',['INSTANCE'], fill_value=-999)
            sdn_depth.standard_name ="sea_floor_depth_below_sea_surface"
            sdn_depth.units='meters'
            sdn_depth.long_name = "Bathymetric depth at "+self.netcdf.featureType+" measurement site"
            sdn_depth.sdn_parameter_urn = "SDN:P01::MBANZZZZ"
            sdn_depth.sdn_parameter_name = "Sea-floor depth (below instantaneous sea level) {bathymetric depth} in the water body"
            sdn_depth.sdn_uom_urn = "SDN:P06::ULAA"
            sdn_depth.sdn_uom_name = "Metres"
            # ONLY IN MARINE ENVIRONMENT !!!!
            evfr['elevation']=evfr['elevation']*-1
            evfr['elevation'].fillna(value=-999, inplace=True)
            sdn_depth[:]=evfr['elevation'].values
            #pangaeas EDMO code            
            sdn_edmo_code=self.netcdf.createVariable('SDN_EDMO_CODE','int',['INSTANCE'])
            sdn_edmo_code.long_name='European Directory of Marine Organisations code for the CDI partner'
            sdn_edmo_code[:]=[3234]*evcnt
            #the mandatory cdi id
            sdn_cdi_id=self.netcdf.createVariable('SDN_LOCAL_CDI_ID','S1',(u'INSTANCE',u'STRING'+str(MaxStrLen['cdi_id'])))
            if self.pandataset.topotype=='vertical profile' or self.pandataset.topotype=='profile series':
                sdn_cdi_id.cf_role='profile_id'
            if self.pandataset.topotype=='time series':
                 sdn_cdi_id.cf_role='timeseries_id'
            sdn_cdi_id.long_name='SeaDataNet CDI identifier'
            cdi_ids=[cdi_id]*evcnt
            sdn_cdi_id[:]=stringtochar(np.array(cdi_ids,'S'+str(MaxStrLen['cdi_id'])))
            
            self.setSDNVariablesAndValues(['INSTANCE',maxtype])
            self.file = self.netcdf.close()

            self.logging.append({'SUCCESS':'NetCDF creation successfully finished'})
        except Exception as e:                       
            if self.netcdf is not None:
                self.netcdf.close()
            self.logging.append({'ERROR':'NetCDF creation failed'+str(e)})
            self.PrintException()

    
        
    def createPANNetCDF(self):
        dim=dict()
        nc = None
        try:
            nc = Dataset(self.filelocation+'\\nc'+str(self.pandataset.id)+'_pan.nc','w',diskless=True, format='NETCDF4')
            self.cleanParameterNames()
            self.setMainVariables(nc)
            ##### PANGAEA STYLE ######
            eventGroup= self.pandataset.data.groupby('Event')
            for eventName, eventFrame in eventGroup:
                self.logging.append({'INFO':'Trying to create NetCDF Dimensions and Variables for Event: '+str(eventName)})
                eventFrame.columns = eventFrame.columns.str.replace('/','_')    
                eventFrame.columns = eventFrame.columns.str.replace(r'[\[\]]','', regex=True)
                topotypeOK=False
                depthColumn='Depth_water'
                depthDir='positive'
                datLen=1
                depLen=1
                #latNo=len(eventFrame['Latitude'].unique())
                #lonNo=len(eventFrame['Longitude'].unique())
                if 'Depth_water' in eventFrame.columns:
                    depNo=len(eventFrame[depthColumn].unique())
                else:
                    depNo=0
                if 'Elevation' in eventFrame.columns:
                    elNo=len(eventFrame['Elevation'].unique())
                else:
                    elNo=0
                if depNo >= 1:
                    depLen = len(eventFrame[depthColumn])
                #datNo=len(eventFrame['Date_Time'].unique())
                #checking topotype
                #replace with panDataSet getGeometry
                pangeotype = self.pandataset.getGeometry()
                if pangeotype in ['trajectoryProfile','timeSeries','profile']:
                    try:                               
                        ng=nc.createGroup(eventName.replace('/','-'))                        
                        ng.createDimension('Latitude', 1)
                        ncLatVar=ng.createVariable('Latitude', 'f4','Latitude')
                        ncLatVar[:]=eventFrame['Latitude'].unique()
                        ng.createDimension('Longitude', 1)
                        ncLonVar=ng.createVariable('Longitude', 'f4','Longitude')
                        ncLonVar[:]=eventFrame['Longitude'].unique()
                        ng.createDimension(depthColumn, depLen)
                        ncDepVar=ng.createVariable(depthColumn, 'f4',depthColumn)
                        if self.pandataset.topotype=='time series':
                            ncDepVar[:]=eventFrame[depthColumn].unique()
                        else:
                            ncDepVar[:]=eventFrame[depthColumn].values
                        ng.createDimension('Date_Time', datLen)
                        ncDatVar=ng.createVariable('Date_Time', 'f8','Date_Time')
                        ncDatVar.long_name='time'
                        ncDatVar.units=self.time_units
                        if self.pandataset.topotype=='time series':
                            ncDatVar[:]=date2num(eventFrame['Date_Time'].dt.to_pydatetime(),units=self.time_units,calendar='standard') 
                        else:
                            ncDatVar[:]=date2num(eventFrame['Date_Time'].iloc[0].to_pydatetime(),units=self.time_units,calendar='standard') 
                    except Exception as e:
                        self.logging.append({'ERROR':'NetCDF grouping failed '+str(e)})
                        self.PrintException()
                        break
                    
                    for pk, p in self.pandataset.params.items():
                        ncvarName=pk.replace('/','_')
                        ncvarName=re.sub('(?![a-zA-Z0-9_]|{MUTF8})([^\x00-\x1F/\x7F-\xFF]|{MUTF8})','_',ncvarName)
                        #ncvarName=re.sub(r'[\[\]]','',ncvarName)
                        if ncvarName not in list(ng.dimensions.keys()) and ncvarName!='Event':
                            if p.type=='numeric':
                                try:
                                    if not eventFrame[ncvarName].isnull().all():
                                        ncVar=ng.createVariable(ncvarName,'f4',['Latitude','Longitude',depthColumn,'Date_Time'])
                                        if p.synonym['CF']!=None:
                                            ncVar.standard_name=p.synonym['CF']
                                        ncVar.long_name=p.name
                                        if p.unit!=None:
                                            ncVar.units=p.unit
                                        else:
                                            ncVar.units='1'
                                        ncVar[:]=eventFrame[ncvarName].values
                                except Exception as e:
                                    self.logging.append({'ERROR':'NetCDF Variable creation failed for Event: '+eventName+' Param: '+ncvarName+', ERROR: '+str(e)})
                                    #print(eventFrame[ncvarName])
                                    self.PrintException()
                                    break
                else:
                    self.logging.append({'ERROR':'NetCDF Variable creation failed: Topotype could not be verified, Dataset was probably wrongly labelled as '+str(self.pandataset.topotype)})
            self.file = nc.close()
            self.logging.append({'SUCCESS': 'PAN style NetCDF created'})
        except Exception as e:                       
            if nc is not None:
                nc.close()   
            self.logging.append({'ERROR':'NetCDF creation failed'+str(e)})




"""
pangaeapy is a package allowing to download and analyse metadata
as well as data from tabular PANGAEA (https://www.pangaea.de) datasets.
"""
from pangaeapy.pandataset import PanAuthor,PanEvent,PanParam,PanDataSet
from pangaeapy.panquery import PanQuery


# -*- coding: utf-8 -*-
"""
Created on Tue Aug 21 13:31:30 2018

@author: Robert Huber
@author: Markus Stocker
@author: Egor Gordeev
@author: Aarthi Balamurugan
"""
import time
import datetime

import requests
import pandas as pd
import numpy as np
import lxml.etree as ET
import re
import io
import os
import textwrap

import logging, logging.handlers

import pickle
from pangaeapy.exporter.pan_netcdf_exporter import PanNetCDFExporter
from pangaeapy.exporter.pan_frictionless_exporter import PanFrictionlessExporter
from pangaeapy.exporter.pan_dwca_exporter import PanDarwinCoreAchiveExporter
class PanProject:
    """PANGAEA Project Class
    This class creates objects which contain the project context information for each dataset
    
    Parameters
    ----------
    acronym : str
        The project acronym
    title : str
        The full title of the project
    URL : str
        The project website
    awardURI : str
        The unique identifier of the award e.g. a CORDIS URI
    id : int
        internal project id

    
    
    Attributes
    ----------
    acronym : str
        The project acronym
    title : str
        The full title of the project
    URL : str
        The project website
    awardURI : str
        The unique identifier of the award e.g. a CORDIS URI
    id : int
        internal project id
    
    """
    def __init__(self,label, name,URL=None, awardURI=None, id=None):
        self.label=label
        self.name=name
        self.URL=URL
        self.awardURI=awardURI
        self.id = id
		
class PanLicence:
	"""PANGAEA Licence Class
	This class contains the licence information for each dataset
    """
	
	def __init__(self,label, name, URI=None):
		self.label=label
		self.name=name
		self.URI=URI
		
class PanAuthor:
    """PANGAEA Author Class.
    A simple helper class to declare 'author' objects which are associated as part of the metadata of a given PANGAEA dataset object
    
    Parameters
    ----------
    lastname : str
        The author's first name
    firstname : str
        The authors's last name
    ORCID : str
        The unique ORCID identifier assigned by orcid.org
    id : int
        The PANGAEA internal id
    
    Attributes
    ----------
    lastname : str
        The author's first name
    firstname : str
        The authors's last name
    fullname : str
        Combination of lastname, firstname. This attribute is created by the constructor
    ORCID : str
        The unique ORCID identifier assigned by orcid.org
    id : int
        The PANGAEA internal id
    """
    def __init__(self,lastname, firstname=None, orcid=None, id=None, affiliations=None):
        self.lastname=lastname
        self.firstname=firstname
        self.fullname=self.lastname
        self.ORCID=orcid
        self.id = id
        if affiliations:
            self.affiliations = affiliations
        if firstname!=None and  firstname!='':
            self.fullname+=', '+self.firstname

class PanEvent:
    """ PANGAEA Event Class.
    An Event can be regarded as a named entity which is defined by the usage of a distinct method or device at a distinct location during a given time interval for scientific purposes.
    More infos on PANGAEA's Evenmts can be found here: https://wiki.pangaea.de/wiki/Event
    
    Parameters
    ----------
    label : str
        A label which is used to name the event
    latitude : float
        The latitude of the event location
    longitude : float
        The longitude of the event location
    elevation : float
        The elevation (relative to sea level) of the event location
    datetime : str
        The date and time of the event in ´%Y/%m/%dT%H:%M:%S´ format
    device : str
        The device which was used during the event
        
    Attributes
    ----------
    label : str
        A label which is used to name the event
    latitude : float
        The start latitude of the event location
    longitude : float
        The start longitude of the event location
    latitude2 : float
        The end latitude of the event location
    longitude2 : float
        The end longitude of the event location
    elevation : float
        The elevation (relative to sea level) of the event location
    datetime : str
        The start date and time of the event in ´%Y/%m/%dT%H:%M:%S´ format
    datetime2 : str
        The end date and time of the event in ´%Y/%m/%dT%H:%M:%S´ format
    device : str
        The device which was used during the event
    basis : PanBasis
        The basis or platform which was used during the event e.g. a ship
    location : str
        The location of the event
    campaign : PanCampaign
        The campaign during which the event was performed
    """
    def __init__(self, label, latitude=None, longitude=None, latitude2=None, longitude2=None, elevation=None, datetime=None, datetime2=None, device=None, basis=None, location=None, campaign=None, id = None, deviceid = None):
        self.label = label
        self.id = id
        if latitude !=None:
            self.latitude=float(latitude)
        else:
            self.latitude=None
        if longitude !=None:
            self.longitude=float(longitude)
        else:
            self.longitude=None
        if latitude2 !=None:
            self.latitude2=float(latitude2)
        else:
            self.latitude2=None
        if longitude2 !=None:
            self.longitude2=float(longitude2)
        else:
            self.longitude2=None
        if elevation !=None:
            self.elevation=float(elevation)
        else:
            self.elevation=None
        self.device = device
        self.deviceid = deviceid
        self.basis = basis
        # -- NEED TO CARE ABOUT datetime2!!!
        self.datetime=datetime
        self.datetime2=datetime2
        self.location=location
        self.campaign=campaign
        
class PanBasis:
    """PANGAEA Basis class
	The basis or platform which was used during the event e.g. a ship
	
	name : str
        The name of the platform
    URI : str
        The platform URI
    callSign : str
        A unique identifier of the platform(alphabet)
    IMOnumber : str
        A seven digit unique identifier for the platform
	"""
    def __init__(self,name=None, URI=None, callSign=None, IMOnumber=None):
        
        self.name=name
        self.URI=URI
        self.callSign=callSign
        self.IMOnumber=IMOnumber
		
class PanCampaign:
    """PANGAEA Campaign class
	The campaign during which the event took place, e.g. a ship cruise or observatory deployment
	
	name : str
        The name of the campaign
    URI : str
        The campaign URI
    start : str
        The start date of the campaign
    end : str
        The end date of the campaign
    startlocation : str
        The start location of the campaign
    endlocation : str
        The end location of the campaign
    BSHID : str
        The BSH ID for the campaign
    expeditionprogram: str
        URL of the campaign report
	"""
    def __init__(self,name=None, URI=None, start=None, end=None, startlocation=None, endlocation=None, BSHID=None, expeditionprogram=None):
        
        self.name=name
        self.URI=URI
        self.start=start
        self.end=end
        self.startlocation=startlocation
        self.endlocation=endlocation
        self.BSHID=BSHID
        self.expeditionprogram=expeditionprogram
	
class PanParam:
    """ PANGAEA Parameter
    Shoud be used to create PANGAEA parameter objects. Parameter is used here to represent 'measured variables'
    
    Parameters
    ----------
    id : int
        the identifier for the parameter
    name : str
        A long name or title used for the parameter
    shortName : str
        A short name or label to identify the parameter
    param_type : str
        indicates the data type of the parameter (string, numeric, datetime etc..)
    source : str
        defines the category or source for a parameter (e.g. geocode, data, event)... very PANGAEA specific ;)
    unit : str
        the unit of measurement used with this parameter (e.g. m/s, kg etc..)
    terms : dict
        a dictionary containing al related terms for this parameter structure:{term:id}
    comment : str
        an optional comment explaining some details of the parameter
    PI : dict
        the responsible PI name of the parameter
    dataseries : int
        the dataseries (column id)
    colno : int
        the column number
    methodid : int
        the id of the used method or device


    
    Attributes
    ----------
    id : int
        the identifier for the parameter
    name : str
        A long name or title used for the parameter
    shortName : str
        A short name or label to identify the parameter
    synonym : dict
        A dictionary of synonyms for the parameter whcih e.g. is used by other archives or communities.
        The dict key indicates the namespace (possible values currently are CF and OS)
    type : str
        indicates the data type of the parameter (string, numeric, datetime etc..)
    source : str
        defines the category or source for a parameter (e.g. geocode, data, event)... very PANGAEA specific ;)
    unit : str
        the unit of measurement used with this parameter (e.g. m/s, kg etc..)
    format: str
        the number format string given by PANGAEA e.g ##.000 which defines the displayed precision of the number
    terms : dict
        a dictionary containing al related terms for this parameter {term:id}
    comment : str
        an optional comment explaining some details of the parameter
    PI : dict
        the responsible PI name of the parameter
    dataseries : int
        the dataseries (column id)
    colno : int
        the column number
    methodid : int
        the id of the used method or device
        
    
    
    """
    def __init__(self, id, name, shortName, param_type, source, unit=None, unit_id = None, format=None, terms =[], comment=None, PI = dict(), dataseries = None, colno = None, methodid = None):
        self.id=id
        self.methodid = methodid
        self.name=name
        self.shortName=shortName
        # Synonym namespace dict predefined keys are CF: CF variables (), OS:OceanSites, SD:SeaDataNet abbreviations (TEMP, PSAL etc..)
        ns=('CF','OS','SD')
        self.synonym=dict.fromkeys(ns)
        self.type=param_type
        self.source=source
        self.unit=unit
        self.unit_id = unit_id
        self.format=format
        self.terms =terms
        self.comment = comment
        self.PI = PI
        self.dataseries = dataseries
        self.colno = colno

    def addSynonym(self,ns, name, uri=None, id=None, unit=None, unit_id = None):
        """
        Creates a new synonym for a parameter which is valid within the given name space. Synonyms are stored in the synonym attribute which is a dictionary
        
        Parameters
        ----------
        name : str
            the name of the synonym
        id : str
            the internal ID of the term used at its auhority ontology
        uri : str
            the URI (ideally actionable) leading to its auhority ontology entry/web page
        ns : str
            the namespace indicator for the sysnonym
        unit : str
            in case another unit expression is used for the synonym parameter within its namespace
        unit_id : str
            the id, UI or URN for the unit
        """
        self.synonym[ns]={'name':name,'id':id, 'uri':uri, 'unit':unit,'unit_id':unit_id}
    
class PanDataSet:
    """ PANGAEA DataSet
    The PANGAEA PanDataSet class enables the creation of objects which hold the necessary information, including data as well as metadata, to analyse a given PANGAEA dataset.
    
    Parameters
    ----------
    id : str
        The identifier of a PANGAEA dataset. An integer number or a DOI is accepted here
    deleteFlag : str
        in case quality flags are avialable, this parameter defines a flag for which data should not be included in the data dataFrame.
        Possible values are listed here: https://wiki.pangaea.de/wiki/Quality_flag
    enable_cache : boolean
        If set to True, PanDataSet objects are cached as pickle files on the local home directory within a directory called 'pangaeapy_cache' in order to avoid unnecessary downloads.
    include_data : boolean
        determines if data table is downloaded and added to the self.data dataframe. If you are interested in metadata only set this to False
    expand_terms : boolean
        indicates if found ontology terms for parameters shall be expanded, i.p. add their hierarchy terms / classification
        
    Attributes
    ----------
    id : str
        The identifier of a PANGAEA dataset. An integer number or a DOI is accepted here
    uri : str
        The PANGAEA DOI (alternative label)
    doi : str
        The PANGAEA DOI
    title : str
        The title of the dataset
    abstract: str
        the abstract or summary of the dataset
    year : int
        The publication year of the dataset
    authors : list of PanAuthor
        a list containing the PanAuthot objects (author info) of the dataset
    citation : str
        the full citation of the dataset including e.g. author, year, title etc..
    params : list of PanParam
        a list of all PanParam objects (the parameters) used in this dataset
    parameters : list of PanParam
        synonym for self.params
    events : list of PanEvent
        a list of all PanEvent objects (the events) used in this dataset
    projects : list of PanProject
        a list containing the PanProjects objects referenced by this dataset
    mintimeextent : str
        a string containing the min time of data set extent
    maxtimeextent : str
        a string containing the max time of data set extent
    data : pandas.DataFrame
        a pandas dataframe holding all the data
    loginstatus : str
        a label which indicates if the data set is protected or not default value: 'unrestricted'            
    isParent : boolean
        indicates if this dataset is a parent data set within a collection of child data sets
    children : list
        a list of DOIs of all child data sets in case the data set is a parent data set
	moratorium : str
		a label which provides the date until the dataset is under moratorium
	datastatus : str
		a label which provides the detail about the status of the dataset whether it is published or in review or deleted
	registrystatus : str
		a string which indicates the registration status of a dataset
	licence : PanLicence
	    a licence object, usually creative commons
	auth_token : str

    """
    def __init__(self, id=None,paramlist=None, deleteFlag='', enable_cache=False, include_data=True, expand_terms=False, auth_token = None):
        self.module_dir = os.path.dirname(os.path.dirname(__file__))
        self.id = None
        self.logging = []
        self.logger = logging.getLogger('dataset')
        ### The constructor allows the initialisation of a PANGAEA dataset object either by using an integer dataset id or a DOI
        self.setID(id)
        self.ns= {'md':'http://www.pangaea.de/MetaData'}        
        # Mapping should be moved to e.g netCDF class/module??
        #moddir = os.path.dirname(os.path.abspath(__file__))
        #self.CFmapping=pd.read_csv(moddir+'\\PANGAEA_CF_mapping.txt',delimiter='\t',index_col='ID')
        self.cache=enable_cache
        self.uri = self.doi = '' #the doi
        self.isParent=False
        self.params=dict()
        self.parameters = self.params
        self.defaultparams=['Latitude','Longitude','Event','Elevation','Date/Time']
        self.paramlist=paramlist
        self.paramlist_index=[]
        self.events=[]
        self.projects=[]
        self.licence=None
        #allowed geocodes for netcdf generation which are used as xarray dimensions not needed in the moment
        self._geocodes={1599:'Date_Time',1600:'Latitude',1601:'Longitude',1619:'Depth water'}
        self.data = pd.DataFrame()
        self.qcdata = pd.DataFrame()
        self.title=None
        self.abstract = None
        self.moratorium=None
        self.curationlevel=None;
        self.processinglevel=None;
        self.datastatus=None
        self.registrystatus=None
        self.citation=None
        self.year=None
        self.date=None
        self.mintimeextent=None
        self.maxtimeextent=None
        self.geometryextent = {}

        self.topotype = None
        self.authors=[]
        #replacing error list

        self.loginstatus='unrestricted'
        self.allowNetCDF=True        
        self.eventInMatrix=False
        self.deleteFlag=deleteFlag
        self.children=[]
        self.include_data=include_data
        self.expand_terms = expand_terms
        self.lastupdate = None
        self.metaxml = None
        self.auth_token = auth_token

        #no symbol = valid(default)
        #? = questionable(?0.345)
        #/ = not valid( / 23.56)
        #* = unknown(*0.999)
        # = individual definition (#999)
        #self.logger.info('Test')
        self.quality_flags={'ok':'valid','?':'questionable','/':'not_valid','*':'unknown'}
        self.quality_flag_replace={'ok':0,'?':1,'/':2,'*':3}
        if self.id != None:
            gotData=False
            #setting up the chache directory in the users home folder
            homedir = os.path.expanduser("~")
            self.cachedir = os.path.join(homedir,'pangaeapy_cache')
            if not os.path.exists(self.cachedir):
                os.makedirs(self.cachedir)

            if self.cache==True:
                self.logging.append({'INFO':'Caching activated..trying to load data and metadata from cache'})
                if self.check_pickle():
                    gotData=self.from_pickle()
                else:
                    self.drop_pickle()
                    gotData = False
            else:
                #delete existing cache
                self.drop_pickle()
            if not gotData:        
                #print('trying to load data and metadata from PANGAEA')
                # check if title is already there, otherwise load metadata
                if not self.title:
                    self.setMetadata()
                if (self.loginstatus=='unrestricted' or self.auth_token)  and self.isParent!=True:
                    self.setData()
                    self.defaultparams = [s for s in self.defaultparams if s in self.params.keys()]
                    if self.paramlist!=None:
                        if  len(self.paramlist)!=len(self.paramlist_index):
                            self.logging.append({'WARNING':'Inconsistent number of detected parameters, expected: '+str(len(self.paramlist))+' vs '+str(len(self.paramlist_index))})
                    if self.cache==True:
                        self.to_pickle()
                else:
                    self.logging.append({'WARNING':'Dataset is either restricted or of type "parent"'})
        else:
            self.logging.append({'ERROR':'Dataset id missing, could not initialize PanDataSet object for: '+str(id)})

    def get_pickle_path(self):
        dirs = textwrap.wrap(str(self.id).zfill(8), 2)
        dirpath = os.path.join(self.cachedir,*dirs)
        try:
            os.makedirs(dirpath)
        except Exception as e:
            pass
        return os.path.join(dirpath , str(self.id) + '_data.pik')

    def check_pickle(self, expirydays = 1):
        '''
        Verifies if a cached pickle files needs to be refreshed (reloaded)
        Files are checked after 24 hrs earliest but only updated in case the metadata indicates changes occured

        Parameters
        ----------
        expirydays

        Returns bool
        -------

        '''
        ret = True
        pickle_location = self.get_pickle_path()
        if os.path.exists(pickle_location):
            pickle_time = os.path.getmtime(pickle_location)
            if int(time.time()) - int(pickle_time) >= (expirydays * 86400):
                # afer 24 hrs check if data set has changed so refresh the cached pickle
                self.setMetadata()
                #2016-10-08T05:40:17
                try:
                    lastupdatets = time.mktime(datetime.datetime.strptime(self.lastupdate, "%Y-%m-%dT%H:%M:%S").timetuple())
                    print(int(lastupdatets) , int(pickle_time))
                    if int(lastupdatets) >= int(pickle_time):
                        ret = False
                        self.logging.append(
                            {'INFO': 'Dataset cache expired, refreshing cache'})
                except Exception as e:
                    print(e)
                    ret = False
                    pass
            return ret

    def drop_pickle(self):
        if os.path.exists(self.get_pickle_path()):
            os.remove(self.get_pickle_path())
                
    def from_pickle(self):
        """
        Reads a PanDataSet object from a pickle file

        """
        ret=False
        pickle_path = self.get_pickle_path()
        if os.path.exists(pickle_path):
            try:
                f = open(pickle_path, 'rb')
                tmp_dict = pickle.load(f)
                tmp_dict['logging'] = []
                f.close()         
                self.__dict__.update(tmp_dict)
                self.logging.append({'INFO':'Loading data and metadata from cache: '+str(pickle_path)})
                ret=True
            except:
                self.logging.append({'WARNING':'Loading data and metadata from cache failed'})
                ret=False
        else:
            ret=False
        return ret
                
    def to_pickle(self):
        """
        Writes a PanDataSet object to a pickle file

        """

        f = open(self.get_pickle_path(), 'wb')
        pickle.dump(self.__dict__, f, 2)
        self.logging.append({'INFO': 'Saved cache (pickle) file at: ' + str(self.get_pickle_path())})
        f.close()
        
    
    def setID (self, id):
        """
        Initialize the ID of a data set in case it was not defined in the constructur
        Parameters
        ----------
        id : str
            The identifier of a PANGAEA dataset. An integer number or a DOI is accepted here
        """
        if type(id) == int:
            self.id = id
        elif isinstance(id, str):
            idmatch = re.search(r'10\.1594\/PANGAEA\.([0-9]+)$', id,re.IGNORECASE)
            if idmatch is not None:
                self.id = idmatch[1]
            else:
                self.logging.append({'ERROR': 'Invalid Identifier or DOI: '+str(id)})
                #print('Invalid Identifier')
        else:
            self.logging.append({'ERROR': 'Invalid Identifier or DOI: ' + str(id)})

    def _getIDParts(self, idstr):
        #returns dict extracted from panmd id strings e.g
        #col13.ds10866878.param7387
        ret = dict()
        if isinstance(idstr, str):
            idmatches = re.findall(r"([a-z]+)([0-9]+)", idstr)
            if idmatches:
                ret = dict(idmatches)
        return ret
    
    def _setEvents(self, panXMLEvents):
        """
        Initializes the list of Events from a metadata XML file for a given pangaea dataset. 
        """
        for event in panXMLEvents:
            eventElevation=eventDateTime=eventDateTime2 = None
            eventDevice=eventLabel = None
            eventDeviceID = None
            basis_name= basis_URI = basis_callsign= basis_imonumber = None
            campaign_name= campaign_URI=campaign_start=campaign_end = None
            eventLongitude=eventLatitude=eventLatitude2=eventLongitude2=eventLocation = None
            startlocation =endlocation=BSHID=expeditionprogram = None

            eventID = self._getIDParts(event.get('id')).get('event')

            if event.find('md:elevation',self.ns)!=None:                
                eventElevation=event.find('md:elevation',self.ns).text
            if event.find('md:dateTime',self.ns)!=None:
                eventDateTime= event.find('md:dateTime',self.ns).text
            if event.find('md:dateTime2',self.ns)!=None:
                eventDateTime2= event.find('md:dateTime2',self.ns).text
            if event.find('md:longitude',self.ns)!=None:
                eventLongitude= event.find('md:longitude',self.ns).text
            if event.find('md:latitude',self.ns)!=None:
                eventLatitude= event.find('md:latitude',self.ns).text
            if event.find('md:longitude2',self.ns)!=None:
                eventLongitude2= event.find('md:longitude2',self.ns).text
            if event.find('md:latitude2',self.ns)!=None:
                eventLatitude2= event.find('md:latitude2',self.ns).text
            if event.find('md:label',self.ns)!=None:
                eventLabel= event.find('md:label',self.ns).text
            if event.find('md:location/md:name',self.ns)!=None:
                eventLocation= event.find('md:location/md:name',self.ns).text
            if event.find('md:method/md:name',self.ns)!=None:
                eventDevice= event.find('md:method/md:name',self.ns).text
                eventDeviceID = self._getIDParts(event.find('md:method',self.ns).get('id')).get('method')
            if event.find('md:basis',self.ns)!=None:
                basis= event.find('md:basis',self.ns)
                if basis.find('md:name',self.ns)!=None:
                    basis_name= basis.find('md:name',self.ns).text
                else:
                    basis_name=None
                if basis.find('md:URI',self.ns)!=None:
                    basis_URI= basis.find('md:URI',self.ns).text
                else:
                    basis_URI=None
                if basis.find('md:callSign',self.ns)!=None:
                    basis_callsign= basis.find('md:callSign',self.ns).text
                else:
                    basis_callsign=None
                if basis.find('md:IMOnumber',self.ns)!=None:
                    basis_imonumber= basis.find('md:IMOnumber',self.ns).text
                else:
                    basis_imonumber=None
                eventBasis=PanBasis(basis_name,basis_URI,basis_callsign,basis_imonumber)
            else:
                eventBasis=None
            if event.find("md:campaign", self.ns)!=None:
                campaign=event.find("md:campaign", self.ns)
                if campaign.find('md:name',self.ns)!=None:
                    campaign_name= campaign.find('md:name',self.ns).text
                if campaign.find('md:URI',self.ns)!=None:
                    campaign_URI= campaign.find('md:URI',self.ns).text
                if campaign.find('md:start',self.ns)!=None:
                    campaign_start= campaign.find('md:start',self.ns).text
                if campaign.find('md:end',self.ns)!=None:
                    campaign_end= campaign.find('md:end',self.ns).text
                if campaign.find('md:attribute[@name="Start location"]',self.ns)!=None:
                    startlocation= campaign.find('md:attribute[@name="Start location"]',self.ns).text
                if campaign.find('md:attribute[@name="End location"]',self.ns)!=None:
                    endlocation= campaign.find('md:attribute[@name="End location"]',self.ns).text
                if campaign.find('md:attribute[@name="BSH ID"]',self.ns)!=None:
                    BSHID= campaign.find('md:attribute[@name="BSH ID"]',self.ns).text
                if campaign.find('md:attribute[@name="Expedition Program"]',self.ns)!=None:
                    expeditionprogram= campaign.find('md:attribute[@name="Expedition Program"]',self.ns).text
                else:
                    expeditionprogram=None
                eventCampaign=PanCampaign(campaign_name,campaign_URI,campaign_start,campaign_end,startlocation,endlocation,BSHID,expeditionprogram)
            else:
                eventCampaign=None

            self.events.append(PanEvent(eventLabel, 
                                        eventLatitude, 
                                        eventLongitude,
                                        eventLatitude2,
                                        eventLongitude2,
                                        eventElevation,
                                        eventDateTime,
                                        eventDateTime2,
                                        eventDevice,
                                        eventBasis,									
                                        eventLocation,
                                        eventCampaign,
                                        eventID,
                                        eventDeviceID
                                        ))

    def _setParameters(self, panXMLMatrixColumn):
        """
        Initializes the list of parameter objects from the metadata XML info
        """
        expandedTerms=dict()
        coln=dict()
        if panXMLMatrixColumn!=None:
            panGeoCode=[]
            for matrix in panXMLMatrixColumn:  
                panparCFName=None
                colno = matrix.get('col')
                paramstr=matrix.find("md:parameter", self.ns)
                #panparID=int(self._getID(str(paramstr.get('id'))))
                paramidparts = self._getIDParts(str(paramstr.get('id')))
                panparID = None
                dataseriesID = None
                if paramidparts.get('param'):
                    panparID = int(paramidparts.get('param'))
                elif paramidparts.get('geocode'):
                    panparID = int(paramidparts.get('geocode'))
                if paramidparts.get('ds'):
                    dataseriesID = int(paramidparts.get('ds'))
                panparShortName='';
                if(paramstr.find('md:shortName',self.ns) != None):
                    panparShortName=paramstr.find('md:shortName',self.ns).text
                    panparIndex = panparShortName
                    #Rename duplicate column headers
                    if panparShortName in coln:
                        coln[panparShortName]+=1
                        panparIndex=panparShortName+'_'+str(coln[panparShortName])
                    else:
                        coln[panparShortName]=1
                panparType=matrix.get('type')
                panparUnit=None
                if(paramstr.find('md:unit',self.ns)!=None):
                    panparUnit=paramstr.find('md:unit',self.ns).text
                panparComment=None
                if(matrix.find('md:comment',self.ns)!=None):
                    panparComment=matrix.find('md:comment',self.ns).text
                panparMethodID = None
                if (paramstr.find('md:method', self.ns) != None):
                    panparMethodID = self._getIDParts(paramstr.find('md:method', self.ns).get('id')).get('method')
                panparPI = None
                panparPI_firstname,panparPI_lastname = None, None
                if(matrix.find('md:PI', self.ns) != None):
                    panparPI_ID = self._getIDParts(matrix.find('md:PI', self.ns).get('id')).get('pi')
                    if(matrix.find('md:PI/md:firstName', self.ns) !=None):
                        panparPI_firstname= matrix.find('md:PI/md:firstName', self.ns).text
                    panparPI_lastname = matrix.find('md:PI/md:lastName', self.ns).text
                    panparPI_fullname = ', '.join(filter(None, [panparPI_lastname,panparPI_firstname]))
                    panparPI = {'id':panparPI_ID,'name':panparPI_fullname, 'last_name':panparPI_lastname,'first_name':panparPI_firstname}
                panparFormat=matrix.get('format')
                if panparShortName=='Event':
                    self.eventInMatrix=True
                #if panparID in self.CFmapping.index:
                #    panparCFName=self.CFmapping.at[panparID,'STDNAME']
                #Add information about terms/ontologies used:
                termlist=[]
                for terminfo in paramstr.findall('md:term', self.ns):
                    termid = None
                    termname = None
                    if terminfo.find("md:name", self.ns) != None:
                        termname = terminfo.find("md:name", self.ns).text
                        termidparts = self._getIDParts(str(terminfo.get('id')))
                        if termidparts.get('term'):
                            termid = int(termidparts.get('term'))
                        terminologyid = int(terminfo.get('terminologyId'))
                        termuri = terminfo.get('semanticURI')
                        try:
                            if self.expand_terms:
                                if isinstance(termid,int):
                                    if termid not in expandedTerms:
                                        try:
                                            termr = requests.get('https://ws.pangaea.de/es/pangaea-terms/term/'+str(termid))
                                            termJSON = termr.json()
                                            if termJSON.get('_source'):
                                                expandedTerms[termid] = []
                                                if termJSON['_source'].get('main_topics'):
                                                    if isinstance(termJSON['_source'].get('main_topics'), list):
                                                        expandedTerms[termid].extend(termJSON['_source'].get('main_topics'))
                                                    else:
                                                        expandedTerms[termid].append(termJSON['_source'].get('main_topics'))
                                                if termJSON['_source'].get('topics'):
                                                    if isinstance( termJSON['_source'].get('topics'), list):
                                                        expandedTerms[termid].extend(termJSON['_source'].get('topics'))
                                                    else:
                                                        expandedTerms[termid].append(termJSON['_source'].get('topics'))
                                        except Exception as e:
                                            self.logging.append({'WARNING':'Failed loading and parsing PANGAEA Term JSON: '+str(e)})
                            if self.expand_terms:
                                if expandedTerms.get(termid):
                                    classification =expandedTerms.get(termid)
                                else:
                                    classification = []
                                termlist.append({'id':termid,'name': str(termname),'semantic_uri':termuri, 'ontology':terminologyid,'classification':classification})
                            else:
                                termlist.append({'id':termid,'name': str(termname),'semantic_uri':termuri, 'ontology':terminologyid})
                        except Exception as e:
                            self.logging.append({'WARNING':'Failed to expand terms '+(str(e))})
                self.params[panparIndex]=PanParam(id=panparID,name=paramstr.find('md:name',self.ns).text,shortName=panparShortName,param_type=panparType,source=matrix.get('source'),unit=panparUnit,format=panparFormat,terms=termlist, comment=panparComment,PI =panparPI, dataseries = dataseriesID, colno = colno, methodid = panparMethodID)
                self.parameters = self.params
                if panparType=='geocode':
                    if panparShortName in panGeoCode:
                        self.allowNetCDF = False
                        self.logging.append({'WARNING': 'Data set contains duplicate Geocodes'})
                    else:
                        panGeoCode.append(panparShortName)

        
    def getEventsAsFrame(self):
        """
        For more convenient handling of event info, this method returns a dataframe containing all events with their attributes as columns
        Please note that this version just takes campaign names, not other campaign attributes
        """
        df=pd.DataFrame()
        try:
            df = pd.DataFrame([ev.__dict__ for ev in self.events ])
            df['campaign'] = df['campaign'].apply(lambda x: x.name)
        except:
            pass
        return df
        
    def setData(self, addEventColumns=True):
        """
        This method populates the data DataFrame with data from a PANGAEA dataset.
        In addition to the data given in the tabular ASCII file delivered by PANGAEA.
        
        
        Parameters:
        -----------
        addEventColumns : boolean
            In case Latitude, Longititude, Elevation, Date/Time and Event are not given in the ASCII matrix, which sometimes is possible in single Event datasets, 
            the setData could add these columns to the dataframe using the information given in the metadata for Event. Default is 'True'

        """

        # converting list of parameters` short names (from user input) to the list of parameters` indexes
        # the list of parameters` indexes is an argument for pd.read_csv
        if self.paramlist!=None:
            self.paramlist += self.defaultparams
            for parameter in self.paramlist:
                iter=0
                for shortName in self.params.keys():
                    if parameter==shortName:
                        self.paramlist_index.append(iter)
                    iter+=1
            if len(self.paramlist)!=len(self.paramlist_index):
                self.logging.append({'WARNING': 'Error entering parameters`short names, inconsitent number of parameters'})
        else:
            self.paramlist_index=None
        if self.include_data==True:
            try:
                dataURL="https://doi.pangaea.de/10.1594/PANGAEA."+str(self.id)
                requestheader = {'Accept': 'text/tab-separated-values'}
                if self.auth_token:
                    requestheader['Authorization'] = 'Bearer '+ str(self.auth_token)
                dataResponse = requests.get(dataURL, headers = requestheader)
                panDataTxt= dataResponse.text
                if int(dataResponse.status_code) == 200:
                    if 'text' in str(dataResponse.headers.get('Content-Type')):
                        panData = re.sub(r"/\*(.*)\*/", "", panDataTxt, 1, re.DOTALL).strip()
                        #Read in PANGAEA Data
                        self.data = pd.read_csv(io.StringIO(panData), index_col=False , on_bad_lines='skip',sep=u'\t',usecols=self.paramlist_index,names=list(self.params.keys()),skiprows=[0])
                        # add geocode/dimension columns from Event

                        #if addEventColumns==True and self.topotype!="not specified":
                        if addEventColumns==True:
                            if len(self.events)==1:
                                if 'Event' not in self.data.columns:
                                    self.data['Event']=self.events[0].label
                                    self.params['Event']=PanParam(0,'Event','Event','string','data',None)
                            if len(self.events)>=1:
                                addEvLat=addEvLon=addEvEle=addEvDat=False
                                if 'Event' in self.data.columns:
                                    if 'Latitude' not in self.data.columns:
                                        addEvLat=True
                                        self.data['Latitude']=np.nan
                                        self.params['Latitude']=PanParam(1600,'Latitude','Latitude','numeric','event','deg')
                                    if 'Longitude' not in self.data.columns:
                                        addEvLon=True
                                        self.data['Longitude']=np.nan
                                        self.params['Longitude']=PanParam(1601,'Longitude','Longitude','numeric','event','deg')
                                    if 'Elevation' not in self.data.columns:
                                        addEvEle=True
                                        self.data['Elevation']=np.nan
                                        self.params['Elevation']=PanParam(8128,'Elevation','Elevation','numeric','event','m')
                                    if 'Date/Time' not in self.data.columns:
                                        addEvDat=True
                                        self.data['Date/Time']=np.nan
                                        self.params['Date/Time']=PanParam(1599,'Date/Time','Date/Time','numeric','event','')
                                    for iev,pevent in enumerate(self.events):
                                        if pevent.latitude is not None and addEvLat==True:
                                            self.data.loc[(self.data['Event']== pevent.label) & (self.data['Latitude'].isnull()),['Latitude']]=self.events[iev].latitude
                                        if pevent.longitude is not None and addEvLon:
                                            self.data.loc[(self.data['Event']== pevent.label) & (self.data['Longitude'].isnull()),['Longitude']]=self.events[iev].longitude
                                        if pevent.elevation is not None and addEvEle:
                                            self.data.loc[(self.data['Event']== pevent.label) & (self.data['Elevation'].isnull()),['Elevation']]=self.events[iev].elevation
                                        if pevent.datetime is not None and addEvDat:
                                            self.data.loc[(self.data['Event']== pevent.label) & (self.data['Date/Time'].isnull()),['Date/Time']]=str(self.events[iev].datetime)
                        # -- delete values with given QC flags
                        if self.deleteFlag!='':
                            if self.deleteFlag=='?' or self.deleteFlag=='*':
                                self.deleteFlag="\\"+self.deleteFlag
                            self.data.replace(regex=r'^'+self.deleteFlag+'{1}.*',value='',inplace=True)

                        # --- Delete empty columns
                        self.data=self.data.dropna(axis=1, how='all')
                        # --- Preserve QC Flags in self.qcdata DataDrame and
                        # --- Replace Quality Flags for numeric columns


                        for paramcolumn in list(self.params.keys()):
                            if paramcolumn not in self.data.columns:
                                del self.params[paramcolumn]

                        self.setQCDataFrame()
                        self.data.replace(regex=r'^[\?/\*#\<\>]',value='',inplace=True)

                        # --- Adjust Column Data Types
                        self.data = self.data.apply(pd.to_numeric, errors='ignore')
                        if 'Date/Time' in self.data.columns:
                            self.data['Date/Time'] = pd.to_datetime(self.data['Date/Time'], format='%Y/%m/%dT%H:%M:%S')
                    else:
                        self.logging.append({'WARNING': 'Dataset seems to be a binary file which cannot be handled by pangaeapy'})
                elif int(dataResponse.status_code) == 401:
                    if self.auth_token:
                        self.logging.append({'WARNING': 'Data access failed, invalid auth token'})
                    else:
                        self.logging.append({'WARNING': 'Data access failed, authorisation failed '})
                else:
                    self.logging.append({'WARNING': 'Data access failed, response code '+(str(dataResponse.status_code))})
            except Exception as e:
                self.logging.append({'ERROR':'Loading data failed, reason: '+str(e)})

    def setQCDataFrame(self):
        tmp_qc_column_list = []
        try:
            for paramcolumn in list(self.params.keys()):
                if self.params[paramcolumn].type in ['numeric', 'datetime']:
                    tmp_qc_series = self.data[paramcolumn].astype(
                        str).str.extract(r'(^[\*/\?])?(.+)')[0]
                    #self.qcdata[paramcolumn] = self.data[paramcolumn].astype(
                    #    str).str.extract(r'(^[\*/\?])?(.+)')[0]
                    #self.qcdata[paramcolumn].fillna(value='ok', inplace=True)
                    tmp_qc_series.replace(to_replace=self.quality_flag_replace, inplace=True)
                    tmp_qc_column_list.append(tmp_qc_series)
                    #self.qcdata[paramcolumn].replace(to_replace=self.quality_flag_replace, inplace=True)
            self.qcdata = pd.concat(tmp_qc_column_list, axis=1)
            self.qcdata = self.qcdata.dropna(how='all')
            self.qcdata.fillna(0,inplace=True)
        except Exception as e:
            self.logging.append({'WARNING': 'Could not create QC flag dataframe'})

    def addQCParamsAndColumns(self, qc_suffix='_QC', excludeColumns=[]):
        #self.data.replace(regex=r'^[\?/\*#\<\>]', value='', inplace=True)
        if excludeColumns:
            joincolumns = list(set(self.data.columns) & set(set(self.qcdata.columns) - set(excludeColumns)))
        else:
            joincolumns = self.data.columns
        self.data = self.data.join(self.qcdata,rsuffix=qc_suffix)
        for paramcolumn in self.qcdata[joincolumns]:
            if self.params[paramcolumn].source == 'data':
                ptype = 'qc'
            else:
                # geocodeqc
                ptype = 'gqc'
            self.data[paramcolumn + qc_suffix].fillna(0)
            self.params[paramcolumn + qc_suffix] = PanParam(self.params[paramcolumn].id + 1000000000,
                                                            self.params[
                                                                paramcolumn].name + qc_suffix,
                                                            self.params[
                                                                paramcolumn].shortName + qc_suffix,
                                                            source='pangaeapy', param_type=ptype)

    def _setCitation(self):
        citationURL="https://doi.pangaea.de/10.1594/PANGAEA."+str(self.id)
        cheaders = {'Accept':'text/x-bibliography'}
        if self.auth_token:
            cheaders['Authorization'] = 'Bearer ' + str(self.auth_token)
        r=requests.get(citationURL, headers=cheaders)
        if r.status_code!=404:
            self.citation=r.text
        else:
            self.logging.append({'WARNING':'Could not retrieve citation info from PANGAEA'})
        
    def setMetadata(self):
        """
        The method initializes the metadata of the PanDataSet object using the information of a PANGAEA metadata XML file.
        
        """
        metaDataURL="https://doi.pangaea.de/10.1594/PANGAEA."+str(self.id)
        mheaders = {'Accept':'application/vnd.pangaea.metadata+xml'}
        if self.auth_token:
            mheaders['Authorization'] = 'Bearer ' + str(self.auth_token)
        r=requests.get(metaDataURL, headers=mheaders)
        if r.status_code==429:
            self.logging.append({'WARNING':'Received too many requests error (429)...'})
            sleeptime = 1
            if r.headers.get('retry-after'):
                sleeptime = r.headers.get('retry-after')
            if int(sleeptime) < 1:
                sleeptime=1
            time.sleep(int(sleeptime))
            r = requests.get(metaDataURL)
            self.logging.append({'INFO':'After repeating request, got status code: '+str(r.status_code)})

        if r.status_code!=404:
            try:
                r.raise_for_status()
                xmlText=r.text
                self.metaxml =xmlText
                xml = ET.fromstring(xmlText.encode())
                #dataset_status = None
                if xml.find('./md:technicalInfo/md:entry[@key="status"]',self.ns) is not None:
                    self.datastatus = xml.find('./md:technicalInfo/md:entry[@key="status"]',self.ns).get('value')
                if self.datastatus not in['deleted', None]:
                    self.loginstatus=xml.find('./md:technicalInfo/md:entry[@key="loginOption"]',self.ns).get('value')
                    if self.loginstatus!='unrestricted':
                        if self.auth_token:
                            self.logging.append({'INFO': 'Trying to load protected dataset using the given auth token'})
                        else:
                            self.logging.append({'WARNING': 'Data set is protected'})
                    if xml.find('./md:technicalInfo/md:entry[@key="lastModified"]', self.ns)!= None:
                        self.lastupdate = xml.find('./md:technicalInfo/md:entry[@key="lastModified"]', self.ns).get('value')
                    hierarchyLevel=xml.find('./md:technicalInfo/md:entry[@key="hierarchyLevel"]',self.ns)
                    if hierarchyLevel!=None:
                        if hierarchyLevel.get('value')=='parent':
                            self.logging.append({'WARNING':'Data set is of type parent, please select one of its child datasets'})
                            self.isParent=True
                            self._setChildren()
                    self.title=xml.find("./md:citation/md:title", self.ns).text
                    if xml.find("./md:abstract", self.ns)!=None:
                        self.abstract = xml.find("./md:abstract", self.ns).text
                    self.registrystatus=xml.find('./md:technicalInfo/md:entry[@key="DOIRegistryStatus"]',self.ns).get('value')
                    if xml.find('./md:technicalInfo/md:entry[@key="moratoriumUntil"]',self.ns) != None:
                        self.moratorium=xml.find('./md:technicalInfo/md:entry[@key="moratoriumUntil"]',self.ns).get('value')
                    if xml.find('./md:status/md:curationLevel/md:name',self.ns)!=None:
                        self.curationlevel= xml.find('./md:status/md:curationLevel/md:name',self.ns).text
                    if xml.find('./md:status/md:processingLevel/md:name',self.ns)!=None:
                        self.processinglevel= xml.find('./md:status/md:processingLevel/md:name',self.ns).text
                    self.year=xml.find("./md:citation/md:year", self.ns).text
                    self.date=xml.find("./md:citation/md:dateTime", self.ns).text
                    if self.lastupdate == None:
                        self.lastupdate = self.date
                    self.doi=self.uri=xml.find("./md:citation/md:URI", self.ns).text
                    #extent
                    if xml.find("./md:extent/md:temporal/md:minDateTime", self.ns)!=None:
                        self.mintimeextent=xml.find("./md:extent/md:temporal/md:minDateTime", self.ns).text
                    if xml.find("./md:extent/md:temporal/md:maxDateTime", self.ns)!=None:
                        self.maxtimeextent=xml.find("./md:extent/md:temporal/md:maxDateTime", self.ns).text
                    if xml.find("./md:extent/md:geographic/md:westBoundLongitude", self.ns) != None:
                        self.geometryextent['westBoundLongitude'] = xml.find("./md:extent/md:geographic/md:westBoundLongitude",
                                                           self.ns).text
                    if xml.find("./md:extent/md:geographic/md:eastBoundLongitude", self.ns) != None:
                        self.geometryextent['eastBoundLongitude'] = xml.find("./md:extent/md:geographic/md:eastBoundLongitude",
                                                           self.ns).text
                    if xml.find("./md:extent/md:geographic/md:southBoundLatitude", self.ns) != None:
                        self.geometryextent['southBoundLatitude'] = xml.find("./md:extent/md:geographic/md:southBoundLatitude",
                                                           self.ns).text
                    if xml.find("./md:extent/md:geographic/md:northBoundLatitude", self.ns) != None:
                        self.geometryextent['northBoundLatitude'] = xml.find("./md:extent/md:geographic/md:northBoundLatitude",
                                                           self.ns).text
                    if xml.find("./md:extent/md:geographic/md:northBoundLatitude", self.ns) != None:
                        self.geometryextent['northBoundLatitude'] = xml.find("./md:extent/md:geographic/md:northBoundLatitude",
                                                           self.ns).text
                    if xml.find("./md:extent/md:geographic/md:meanLongitude", self.ns) != None:
                        self.geometryextent['meanLongitude'] = xml.find("./md:extent/md:geographic/md:meanLongitude",
                                                           self.ns).text
                    if xml.find("./md:extent/md:geographic/md:meanLatitude", self.ns) != None:
                        self.geometryextent['meanLatitude'] = xml.find("./md:extent/md:geographic/md:meanLatitude",
                                                           self.ns).text

                    topotypeEl=xml.find("./md:extent/md:topoType", self.ns)
                    if topotypeEl!=None:
                        self.topotype=topotypeEl.text
                    else:
                        self.topotype=None
                    for author in xml.findall("./md:citation/md:author", self.ns):
                        lastname=None
                        firstname=None
                        orcid=None
                        if author.find("md:lastName", self.ns)!=None:
                            lastname=author.find("md:lastName", self.ns).text
                        if author.find("md:firstName", self.ns)!=None:
                            firstname=author.find("md:firstName", self.ns).text
                        if author.find("md:orcid", self.ns)!=None:
                            orcid=author.find("md:orcid", self.ns).text
                        #if author.find("md:affiliation", self.ns)!=None:
                        authoraffiliations = []
                        for affiliations in author.findall("md:affiliation", self.ns):
                            afm = re.search(r"\.inst([0-9]+)$",str(affiliations.get('id')))
                            if afm:
                                authoraffiliations.append(afm[1])
                        #print(authoraffiliations)
                        authorid=author.get('id')
                        if authorid:
                            authorid=int(authorid.replace('dataset.author',''))
                        self.authors.append(PanAuthor(lastname, firstname,orcid,authorid, authoraffiliations))
                    for project in xml.findall("./md:project", self.ns):
                        label=None
                        name=None
                        URI=None
                        awardURI=None
                        if project.find("md:label", self.ns)!=None:
                            label=project.find("md:label", self.ns).text
                        if project.find("md:name", self.ns)!=None:
                            name=project.find("md:name", self.ns).text
                        if project.find("md:URI", self.ns)!=None:
                            URI=project.find("md:URI", self.ns).text
                        if project.find("md:award/md:URI", self.ns)!=None:
                            awardURI=project.find("md:award/md:URI", self.ns).text
                        if project.get('id'):
                            projectid=str(project.get('id')).replace('project','')
                        self.projects.append(PanProject(label, name, URI, awardURI,int(projectid)))
                    if xml.find("./md:license",self.ns) != None:
                        license = xml.find("./md:license",self.ns)
                        label=None
                        name=None
                        URI=None
                        if license.find("md:label", self.ns)!=None:
                            label=license.find("md:label", self.ns).text
                        if license.find("md:name", self.ns)!=None:
                            name=license.find("md:name", self.ns).text
                        if license.find("md:URI", self.ns)!=None:
                            URI=license.find("md:URI", self.ns).text
                        self.licence = PanLicence(label, name, URI)

                    panXMLMatrixColumn=xml.findall("./md:matrixColumn", self.ns)
                    self._setParameters(panXMLMatrixColumn)
                    panXMLEvents=xml.findall("./md:event", self.ns)
                    self._setEvents(panXMLEvents)
                else:
                    self.logging.append({'ERROR': 'Dataset is deleted or of unknown status: ' + str(self.datastatus)})
            except requests.exceptions.HTTPError as e:
                self.logging.append({'ERROR': 'Failed to retrieve metadata information: '+str(e)})
            except ET.ParseError as e:
                self.logging.append({'ERROR': 'Failed to parse metadata information: '+str(e)})
            if self.datastatus not in ['deleted', None]:
                self._setCitation()

        else:
            self.logging.append({'ERROR': 'Data set does not exist, 404'})
            self.id = None

    def _setChildren(self):
        childqueryURL="https://www.pangaea.de/advanced/search.php?q=incollection:"+str(self.id)+"&count=1000"
        r = requests.get(childqueryURL)
        if r.status_code != 404:
            s = r.json()
            for p in s['results']: 
                self.children.append(p['URI'])
                #print(p['URI'])
            
    def getGeometry(self):
        """
        Sometimes the topotype attribute has not been set correctly during the curation process. 
        This method returns the real geometry (topographic type) of the dataset based on the x,y,z and t information of the data frame content.
        Still a bit experimental..
        """
        geotype=0
        zgroup=['Latitude','Longitude']
        tgroup=['Latitude','Longitude']
        locgrp=['Latitude','Longitude']
        p=pz=pt=len(self.data.groupby(locgrp))
        t=z=None
        
        if 'Date_Time' in self.data.columns:
            t='Date_Time'
        if 'Depth_water' in self.data.columns:
            z='Depth_water'
        elif 'Depth' in self.data.columns:
            z='Depth'
        elif 'Depth_ice_snow' in self.data.columns:
            z='Depth_ice_snow'
        elif 'Depth_soil' in self.data.columns:
            z='Depth_soil'

        if t!=None:
            tgroup.append(t)
            pt=len(self.data.groupby(tgroup))
        if z!=None:
            zgroup.append(z)
            pz=len(self.data.groupby(zgroup))
        if p==1:
            if pt==1 and pz==1:
                geotype='point'
            elif pt>=1:
                if pz==1 or len(self.events)==1:
                    geotype='timeSeries'
                else: 
                    geotype='timeSeriesStack'
            else:
                geotype='profile'
        else:
            if p==pz:
                geotype='trajectory'
            elif pt>pz:
                geotype='timeSeriesProfile'
            else:
                geotype='trajectoryProfile'
        return geotype


    """
            The method returns translates the parameter object list into a dictionary 
    """
    def getParamDict(self):
        paramdict={'shortName':[],'name':[],'unit':[],'type':[],'format':[]}
        for param_key, param_value in self.params.items():
            paramdict['shortName'].append(param_key)
            paramdict['name'].append(param_value.name)
            paramdict['unit'].append(param_value.unit)
            paramdict['type'].append(param_value.type)
            paramdict['format'].append(param_value.format)
        return paramdict

    """
            The method returns a set of basic information about the PANGAEA dataset
    """
    def info(self):
        print('Citation:')
        print(self.citation)
        print()
        print('Toptype:')
        print(self.topotype)
        print()
        print('Parameters:')
        paramdf=pd.DataFrame(self.getParamDict())
        print(paramdf)
        print()
        print('Data: (first 5 rows)')
        print(self.data.head(5))

    def rename_column(self,old_name, new_name):
        if self.params.get(old_name):
            self.params[new_name] = self.params.pop(old_name)
            self.params[new_name].name = new_name
            self.params[new_name].shortName = new_name
            self.data.rename(columns={old_name:new_name}, inplace = True)
            self.qcdata.rename(columns={old_name: new_name}, inplace=True)


    def to_netcdf(self, filelocation=None, save = True, type='sdn'):
        """
        This method creates a NetCDF file using PANGAEA data. It offers two different flavors: SeaDataNet NetCDF and an
        experimental internal format using NetCDF 4 groups.
        Currently the method only supports simple types such as timeseries and profiles.
        The method created files are named as follows: [PANGAEA ID]_[type].cf

        Parameters:
        -----------
        filelocation : str
            Indicates the location (directory) where the NetCDF file will be saved
        type : str
            This parameter sets the NetCDF profile type. Allowed values are 'sdn' (SeaDataNet) and 'pan' (PANGAEA style)
        save : Boolean
            If the file shall be saved on disk (filelocation or home directory/pan_export by default)
        """
        ret = None
        netcdfexporter = PanNetCDFExporter(self, filelocation=filelocation)
        if type == 'sdn':
            netcdfexporter.renameSDNDimVars()
            netcdfexporter.pandataset.addQCParamsAndColumns(qc_suffix='_SEADATANET_QC',
                                                excludeColumns=['LATITUDE', 'LONGITUDE', 'TIME'])
        ret = netcdfexporter.create(style=type)
        if save:
            netcdfexporter.save()
        return ret

    def to_frictionless(self, filelocation=None, save = True):
        """
        This method creates a frictionless data package (https://specs.frictionlessdata.io/data-package) file using PANGAEA metadata and data.
        A package will be saved as directory
        The method created directories are named as follows: [PANGAEA ID]_frictionless

        Parameters:
        -----------
        filelocation : str
            Indicates the location (directory) where the frictionless file will be saved
        save : Boolean
            If the file shall be saved on disk (filelocation or home directory/pan_export by default)
        """

        frictionless_exporter = PanFrictionlessExporter(self, filelocation)
        ret= frictionless_exporter.create()
        if save:
            frictionless_exporter.save()
        return ret

    def to_dwca(self, save = True):
        """
        This method creates a Darwin Core Archive file using PANGAEA metadata and data.
        A package will be saved as directory
        The method created directories are named as follows: [PANGAEA ID]_dwca

        Parameters:
        -----------
        filelocation : str
            Indicates the location (directory) where the DwC-A file will be saved
        save : Boolean
            If the file shall be saved on disk (filelocation or home directory/pan_export by default)
        """
        dwca_exporter = PanDarwinCoreAchiveExporter(self)
        ret = dwca_exporter.create()
        if save:
            dwca_exporter.save()
        #print(dwca_exporter.logging)
        self.logging.extend(dwca_exporter.logging)
        return ret



import os
import re
from collections import OrderedDict

import lxml.etree
import lxml.etree as et
from pangaeapy.exporter.pan_exporter import PanExporter
from zipfile import ZipFile
from io import BytesIO

class PanDarwinCoreAchiveExporter(PanExporter):

    def __init__(self, *args, **kwargs):
        super(PanDarwinCoreAchiveExporter, self).__init__(*args, **kwargs)

        self.dwcnames = {'Event': 'EventID', 'Latitude': 'decimalLatitude', 'Longitude': 'decimalLongitude',
                    'Date/Time': 'eventDate'}
        self.dwcfields = ['id', 'modified', 'institutionCode', 'CollectionCode', 'datasetID', 'basisOfRecord', 'catalogNumber',
                     'recordedBy', 'eventDate', 'scientificName', 'kingdom', 'geodeticDatum', 'decimalLatitude',
                     'decimalLongitude', 'organismQuantity', 'organismQuantityType']
        #http://vocab.nerc.ac.uk/collection/S11/current/
        self.taxon_lifestages = ['adult','juvenile','larvae','eggs','nauplii','copepodites']
        #http://vocab.nerc.ac.uk/collection/S10/current/
        self.taxon_sex = ['male','female','hermaphrodite']
        self.taxon_attributes = self.taxon_lifestages + self.taxon_sex
        self.taxon_attributes.append('total')
        self.chronostrat_params = [21496, 21497, 21498, 20544, 21197]
        # XYZ zone
        self.biostrat_params = [4491, 15398, 15501, 20543, 21181, 85701, 51065, 57117, 57648, 86368, 89835, 121195,
                           128347, 135504, 170899, 185728]
        # comment
        self.agecomment_params = [643]
        # absolute dated ages
        self.absstrat_params = [2205, 5506, 6167, 6168, 6169, 6170, 70169, 102659, 130805, 145907]
        self.taxonomic_ontologies = [1,2]
        self.taxonomic_coverage = []
        self.known_synonyms = {'Coccolithophoridae':'Coccolithophorida'}

    def set_elevation_column(self):
        ret = 'pos'
        verticaldirection = {'Depth water':'neg', 'Elevation':'pos'}
        for vgeo, vdir in verticaldirection.items():
            if vgeo in self.pandataset.data.columns:
                ret = vdir
                self.dwcnames[vgeo] = 'minimumElevationInMeters'
                self.dwcfields.append('minimumElevationInMeters')
                break
        return ret

    def check_unit(self, unitexpr):
        unitre = '^([#%])(?:\/((?:[0-9]+\s)?(?:[kdcm]?m{1,2}\*{2}[23]|m?l|k?g)))?(?:\/(d|m|y|a|ka|day|week|month|year){1})?$'
        dimension = ''
        istaxonrelated = False
        if unitexpr:
            if isinstance(unitexpr, str):
                try:
                    umatch = re.search(unitre, unitexpr)
                    if umatch:
                        uparts = filter(None, list(umatch.groups()))
                        if umatch[1] is not None:
                            istaxonrelated = True
                            if umatch[1] == '#':
                                dimension = 'individuals'
                            else:
                                dimension = 'percentage'
                            if umatch[2] is not None:
                                if 'l' in umatch[2] or '**3' in umatch[2]:
                                    dimension += ' per volume'
                                elif 'g' in umatch[2]:
                                    dimension += ' per mass'
                                else:
                                    dimension += ' per area'
                            if umatch[3] is not None:
                                if umatch[2] is not None:
                                    dimension += ' and time'
                                else:
                                    dimension += ' per time'
                except Exception as e:
                    self.logging.append({'WARNING': 'Unit check failed: '+str(e)})
        else:
            dimension = 'relative abundance'
            istaxonrelated =True
        return istaxonrelated, dimension

    def get_taxon_columns(self):
        taxoncolumns = OrderedDict()
        taxon_attr_regex = r'(.*?)((?:,\s?)'+str('|(?:,\s?)'.join(self.taxon_attributes))+')$'
        for pkey, param in self.pandataset.params.items():
            # full match of taxon name with parameter only
            # TODO: extend to some adjectives e.g. juvenile, adult etc..
            try:
                for term in param.terms:
                    name_parts = re.split(r',\s?',param.name)
                    taxon_candidate = str(param.name)

                    taxon_attribute = None
                    if len(name_parts) == 2:
                        if name_parts[1] in self.taxon_attributes:
                            taxon_candidate = name_parts[0]
                            taxon_attribute = name_parts[1]
                    if taxon_candidate in self.known_synonyms:
                        taxon_candidate = self.known_synonyms.get(taxon_candidate)


                    # add: #/m3 etc, %/m3 etc
                    is_valid_unit,  dimension = self.check_unit(param.unit)

                    test_taxon = taxon_candidate

                    if taxon_candidate.endswith(' sp.'):
                        test_taxon  = taxon_candidate.replace(' sp.','').strip()
                    if taxon_candidate.endswith(' spp.'):
                        test_taxon  = taxon_candidate.replace(' spp.','').strip()

                    if test_taxon .lower() == str(term.get('name')).lower() and is_valid_unit:
                        if term.get('classification'):
                            if 'Biological Classification' in term.get('classification'):
                                kingdom = list({'Animalia', 'Archaea', 'Bacteria', 'Chromista', 'Fungi', 'Plantae', 'Protozoa',
                                                'Viruses'} & set(term.get('classification')))[0]
                                if kingdom not in self.taxonomic_coverage:
                                    self.taxonomic_coverage.append(kingdom)

                                taxoncolumns[pkey] = {'taxon': taxon_candidate, 'series': param.dataseries, 'author': param.PI,
                                                      'kingdom': kingdom, 'colno': param.colno, 'unit': param.unit,'dimension': dimension}
                                if taxon_attribute:
                                    if taxon_attribute in self.taxon_sex:
                                        taxoncolumns[pkey]['sex'] = taxon_attribute
                                        if 'sex' not in self.dwcfields:
                                            self.dwcfields.append('sex')
                                    if taxon_attribute in self.taxon_lifestages:
                                        taxoncolumns[pkey]['lifestage'] = taxon_attribute
                                        if 'lifeStage' not in self.dwcfields:
                                            self.dwcfields.append('lifeStage')
                    break

            except Exception as e:
                self.logging.append({'WARNING': 'Failed to identify taxonomic information in parameter: '+str(pkey)})
        if len(taxoncolumns) <= 0:
            self.logging.append({'WARNING': 'Could not identify taxonomic information in this dataset'})
        return taxoncolumns

    def get_context_info(self):
        geocontext_params = self.chronostrat_params + self.biostrat_params + self.agecomment_params + self.absstrat_params
       # print(geocontext_params)
        basisofrecord = 'HumanObservation'
        geologicalcontextid = None
        try:
            for param in self.pandataset.params.values():
                if param.id:
                    if int(param.id) in geocontext_params:
                        #basisofrecord = 'FossilSpecimen'
                        geologicalcontextid = 'http://purl.obolibrary.org/obo/ENVO_00002164'
                    if int(param.id) in [1]:
                        #basisofrecord = 'FossilSpecimen'
                        geologicalcontextid = 'http://purl.obolibrary.org/obo/ENVO_00002007'
        except Exception as e:
            self.logging.append({'INFO': 'Failed to identify basisOfRecord based on used geocodes, keep HumanObservation by default'})

        return basisofrecord, geologicalcontextid

    def get_dwca_data(self, taxoncolumns):
        dwcdata = None
        basisofrecord, geologicalcontextid = self.get_context_info()
        selectedcolumns = []
        geocolumns = self.pandataset.defaultparams
        if 'Depth water' in self.pandataset.data.columns:
            geocolumns.append('Depth water')
        if len(taxoncolumns) > 0:
            selectedcolumns.extend(geocolumns)
            selectedcolumns.extend(taxoncolumns.keys())
            taxonframe = self.pandataset.data[selectedcolumns]
            taxonframe = taxonframe.reset_index()
            geocolumns.append('index')
            taxonframe = taxonframe.melt(id_vars=geocolumns, value_vars=list(taxoncolumns.keys()), var_name='Colname',
                                         value_name='organismQuantity')
            taxonframe['index'] += 1
            taxonframe['id'] = taxonframe['index'].astype(str) + '_' + taxonframe['Colname'].apply(
                lambda x: taxoncolumns.get(x).get('colno')).astype(str)
            taxonframe['modified'] = self.pandataset.lastupdate
            taxonframe['institutionCode'] = 'Pangaea'
            doimatch = re.search('(10\.1594/PANGAEA\.[0-9]+)', self.pandataset.doi)
            taxonframe['CollectionCode'] = 'doi:' + str(doimatch[1])
            taxonframe['datasetID'] = self.pandataset.doi
            taxonframe['basisOfRecord'] = basisofrecord
            taxonframe['catalogNumber'] = taxonframe['Colname'].apply(
                lambda x: taxoncolumns.get(x).get('series')).astype(str) + '_' + taxonframe['index'].astype(str)
            taxonframe['recordedBy'] = taxonframe['Colname'].apply(lambda x: None if not taxoncolumns.get(x).get('author') else taxoncolumns.get(x).get('author').get('name'))
            taxonframe['scientificName'] = taxonframe['Colname'].apply(lambda x: taxoncolumns.get(x).get('taxon'))
            taxonframe['geodeticDatum'] = 'WGS84'
            taxonframe['kingdom'] = taxonframe['Colname'].apply(lambda x: taxoncolumns.get(x).get('kingdom'))
            #taxonframe['organismQuantityType'] = 'individuals (' + taxonframe['Colname'].apply(
            #    lambda x: taxoncolumns.get(x).get('unit')).astype(str) + ')'
            taxonframe['organismQuantityType'] = taxonframe['Colname'].apply(lambda x: taxoncolumns.get(x).get('dimension'))
            try:
                if 'sex' in self.dwcfields:
                    taxonframe['sex'] = taxonframe['Colname'].apply(lambda x: taxoncolumns.get(x).get('sex'))

                if 'lifeStage' in self.dwcfields:
                    taxonframe['lifeStage'] = taxonframe['Colname'].apply(lambda x: taxoncolumns.get(x).get('lifestage'))
            except Exception as e:
                print(e)
            replace_dwcnames = {ck: cv  for (ck, cv) in self.dwcnames.items() if ck in taxonframe.columns}
            taxonframe.rename(columns=replace_dwcnames, inplace=True)
            dwcfields= [f for f in self.dwcfields if f in taxonframe.columns]
            if geologicalcontextid:
                taxonframe['geologicalContextID'] = geologicalcontextid
                dwcfields.append('geologicalContextID')
            elevation_direction = self.set_elevation_column()

            taxonframe = taxonframe[dwcfields]

            if elevation_direction == 'neg' and 'minimumElevationInMeters' in taxonframe.columns:
                taxonframe['minimumElevationInMeters'] = taxonframe['minimumElevationInMeters'] * -1

            taxonframe = taxonframe[taxonframe['organismQuantity'].notna()]

            dwcdata = taxonframe.to_csv(index=False,sep='|',line_terminator='\n',date_format ='%Y-%m-%dT%H:%M:%S', encoding='utf-8')
            return dwcdata
        else:
            self.logging.append({'ERROR': 'No taxonomic information identified in dataset, skipping DwC-A ASCII table generation'})
            #print('could not identify taxon counts in this data set')
            return False

    def get_meta_xml(self):
        xml='<archive xmlns="http://rs.tdwg.org/dwc/text/" metadata="eml.xml">' \
            '<core encoding="UTF-8" fieldsTerminatedBy="|" linesTerminatedBy="\n" ' \
            'fieldsEnclosedBy="" ignoreHeaderLines="1" rowType="http://rs.tdwg.org/dwc/terms/Occurrence">' \
            '<files><location>'+str(self.pandataset.id)+'_data.tab</location></files>' \
            '<id index="0"/>'
        index=1
        try:
            for col in self.dwcfields:
                if col != 'id':
                    if str(col) in ['modified']:
                        xml += '<field index="'+str(index)+'" term="http://purl.org/dc/terms/' + str(col) + '" />'
                    else:
                        xml += '<field index="'+str(index)+'" term="http://rs.tdwg.org/dwc/terms/'+str(col)+'" />'
                    index+=1
        except Exception as e:
            self.logging.append({'ERROR': 'Failed to create DwC-A metadata file'})
            xml=None
        if xml:
            xml += '</core></archive>'
            return xml
        else:
            return False


    def get_eml_xml(self):
        ret = False
        if self.pandataset.metaxml:
            try:
                eml_xslt = os.path.join(os.path.dirname(__file__), 'xslt', 'panmd2eml.xslt')
                xslt = et.parse(eml_xslt)
                panxml = et.fromstring(self.pandataset.metaxml.encode())
                transform = et.XSLT(xslt)
                emlxml = transform(panxml)
                coverage = emlxml.find("dataset/coverage")
                if coverage is not None and self.taxonomic_coverage:
                    taxcovelem = lxml.etree.Element('taxonomicCoverage')
                    coverage.append(taxcovelem)
                    for taxcov in self.taxonomic_coverage:
                        taxcovelem.append(lxml.etree.XML('<taxonomicClassification>'
                                                                    '<taxonRankName>kingdom</taxonRankName>'
                                                                    '<taxonRankValue>'+str(taxcov)+'</taxonRankValue>'
                                                                    '</taxonomicClassification>'))
                #coverage.append(lxml.Element('taxonomicCoverage'))
                ret= et.tostring(emlxml, pretty_print=True)
            except Exception as e:
                self.logging.append({'ERROR': 'Failed to perform XSLT transformation to EML: '+str(e)})
        else:
            self.logging.append({'ERROR': 'Failed to perform XSLT transformation to EML, missing panmd XML'})
        return ret

    def verify(self):
        ret = False
        if self.pandataset.id:
            try:
                datacolumns = self.get_taxon_columns()
                if len(datacolumns) > 0:
                    ret = True
            except Exception as e:
                self.logging.append({'ERROR':'DwC-A verification failed: '+str(e)})

        return ret

    def create(self):
        in_memory_zip = False
        if self.pandataset.id:
            try:
                datacolumns = self.get_taxon_columns()
                data = self.get_dwca_data(datacolumns)
                meta = self.get_meta_xml()
                eml = self.get_eml_xml()
                if not any('ERROR' in lg for lg in self.logging):
                    in_memory_zip = BytesIO()
                    zip_file = ZipFile(in_memory_zip, 'w')
                    zip_file.writestr('meta.xml', meta)
                    zip_file.writestr('eml.xml', eml)
                    zip_file.writestr(self.pandataset.id+'_data.tab', data)
                    zip_file.close()
                    in_memory_zip.seek(0)
                    self.file = in_memory_zip
                else:
                    self.logging.append({'ERROR': 'DwC-A Zip file creation failed due to previous errors '})
            except Exception as e:
                self.logging.append({'ERROR':'DwC-A Zip file creation failed: '+str(e)})
        else:
            self.logging.append({'ERROR': 'Not PanDataSet object available to perform the DwC-A export'})

        return in_memory_zip

    def save(self):
        if isinstance(self.file, BytesIO):
            try:
                with open(os.path.join(self.filelocation,str('dwca_pangaea_'+str(self.pandataset.id)+'.zip')),'wb') as f:
                    #print(f.name)
                    f.write(self.file.getbuffer())
                    f.close()
                    self.logging.append({'INFO': 'Saved DwC-A Zip: ' + os.path.join(self.filelocation,str('dwca_pangaea_'+str(self.pandataset.id)+'.zip'))})
                    return True
            except Exception as e:
                self.logging.append({'ERROR': 'Could not save, DwC-A Zip: '+str(e)})
        else:
            self.logging.append({'ERROR':'Could not save, DwC-A Zip file is not a BytesIO'})
            return False


import requests
import re
class PanQuery:
    """PANGAEA Query Class
        This class allows to query PANGAEA and to access query results

        Parameters
        ----------
        query : str
            The query string following the specs at www.pangaea.de
        bbox : set
            The bounding box to define  geographical search constraints following the GeoJSON specs
            example : bbox=(minlon, minlat,  maxlon, maxlat)
        limit : int
            The expected number of search results (max =100)
        offset : int
            The offset of the search to continue results retrieval

        Attributes
        ----------
        totalcount : int
            The number of total search results
        error : str
            In case an error occurs this attribute holds the latest one
        query : str
            The query (see above)
        result : list
            A list of retrieved search results

        """
    PANGAEA_QUERY_URL = 'https://www.pangaea.de/advanced/search.php?q='
    def __init__(self, query, bbox =(), limit=10, offset=0):
        self.totalcount=0
        self.error = None
        self.query = query
        self.result = self._search(self.query, bbox, limit, offset)
        if self.error is not None:
            print('ERROR: '+self.error)


    def _search(self,query, bbox =(), limit=10, offset=0):
        """
                Performs the search request and stores the results in a list
                this method is called by the constructur

        """
        response = []
        panquery = self.PANGAEA_QUERY_URL+query
        try:
            if len(bbox) ==4:
                minlon, minlat,  maxlon, maxlat, = bbox
                panquery += '&maxlat='+str(maxlat)+'&minlon='+str(minlon)+'&maxlon='+str(maxlon)+'&minlat='+str(minlat)
            if offset != 0:
                panquery +='&offset='+str(offset)
            if limit !=10:
                panquery+='&count='+str(limit)

            r = requests.get(panquery)
            res = r.json()
            if r.status_code == 200:
                self.totalcount= res.get('totalCount')
                response = res.get('results')
                position = offset
                for rset in response:
                    if re.search(r'>[0-9]+\sdatasets<', rset.get('html')):
                        rset['type']='parent'
                    else:
                        rset['type']='child'
                    rset['position'] = position
                    position+=1
            elif r.status_code == 500:
                self.error = res.get('error')
            else:
                self.error = 'Request failed: response code: '+str(r.status_code)
        except Exception as e:
            self.error = 'Request failed: '+str(e)
        return response



from pangaeapy.exporter.pan_exporter import PanExporter
import os
import json
class PanPanImportExporter(PanExporter):

    def create_data_string(self):
        return self.pandataset.data.loc[:, ~self.pandataset.data.columns.isin(self.ignorecolumns)].head()

    def create_json_header(self):
        panmetadict = {}
        panmetadict["Title"] =  self.pandataset.title
        panmetadict["DataSetID"] = int(self.pandataset.id)
        panmetadict["Authors"] = []
        panmetadict["Parameter"] = []
        events = []
        eventmethodid = None
        for event in self.pandataset.events:
            events.append({'id':int(event.id),'device':int(event.deviceid)})
        if len(events) == 1:
            panmetadict["EventID"] = events[0]['id']
            eventmethodid = events[0]['device']
        for author in self.pandataset.authors:
            inst1, inst2 = 0,0
            try:
                inst1 = int(author.affiliations[0])
            except:
                pass
            try:
                inst2 = int(author.affiliations[1])
            except:
                pass
            panmetadict["Authors"].append({'ID':author.id, 'InstitutionID':inst1, 'Institution2ID':inst2})
        self.ignorecolumns = []
        for paramk, param in self.pandataset.params.items():
            if param.source not in ['event'] or not param.id:
                if not param.id and param.name=='Event label':
                    param.id = 500000
                paramdict = {'ID':param.id}
                print('Param ID', param.id)

                if param.format:
                    paramdict['Format'] = param.format
                else:
                    paramdict['Format'] = ''
                if param.methodid:
                    methodid = param.methodid
                else:
                    methodid = eventmethodid
                if methodid:
                    paramdict['MethodID'] = methodid
                if param.PI:
                    paramdict['PI_ID'] = param.PI.get('id')
                panmetadict["Parameter"].append(paramdict)
            else:
                self.ignorecolumns.append(paramk)
        print(self.ignorecolumns)
        return json.dumps(panmetadict)

    def create(self):
        panimportstr = self.create_json_header()
        print(self.create_data_string())
        return panimportstr



import shutil
from io import BytesIO
from zipfile import ZipFile

from pangaeapy.exporter.pan_exporter import PanExporter
import os
import json
class PanFrictionlessExporter(PanExporter):

    def get_csv(self):
        csv =''
        #print(self.pandataset.data.head())
        try:
            csv = self.pandataset.data.to_csv(os.path.join(self.filelocation,+'data.csv'), index=False)
        except Exception as e:
            self.logging.append({'ERROR': 'Frictionless CSV creation failed: '+str(e)})

        return csv

    def create_tableschema_json(self):
        schema = {'fields':[]}
        typeconv={'numeric':'number'}
        for k, p in self.pandataset.params.items():
            field={'name':k,'title':p.name}
            #print(k, p.name, p.shortName)
            pantype = p.type
            if pantype in typeconv:
                pantype = typeconv.get(pantype)
            field['type']= pantype
            panunit = p.unit
            if panunit:
                field['unit'] = panunit
            pancomment = p.comment
            if pancomment:
                field['description'] = pancomment
            schema['fields'].append(field)
        return schema

    def get_package_json(self):
        package = {'profile':'tabular-data-package'}
        panauthors = []
        try:
            for author in self.pandataset.authors:
                panauthors.append({'title': author.firstname + ' ' + author.lastname, 'role': 'author'})
            table_schema = self.create_tableschema_json()

            resources=  [{'profile':'tabular-data-resource',
                          'path': self.pandataset.id + '_data.csv',
                          'schema':table_schema}]

            package['name'] = self.pandataset.id + '_metadata'
            package['id'] = self.pandataset.doi
            package['title'] = self.pandataset.title
            if self.pandataset.abstract:
                package['description'] = self.pandataset.abstract
            package['created'] = self.pandataset.date
            package['contributors'] = panauthors
            package['licenses'] = [{'path':self.pandataset.licence.URI, 'name':self.pandataset.licence.label, 'title':self.pandataset.licence.name}]
            package['resources'] =resources
        except Exception as e:
            self.logging.append({'ERROR': 'Frictionless JSON creation failed: '+str(e)})
        return json.dump(package)

    def create(self):
        in_memory_zip = False
        ret = False
        if not self.pandataset.isParent:
            if self.pandataset.loginstatus == 'unrestricted':
                try:
                    in_memory_zip = BytesIO()
                    zip_file = ZipFile(in_memory_zip, 'w')
                    package = self.get_package_json()
                    csv = self.get_csv()
                    zip_file.writestr(self.pandataset.id + '_data.csv', csv)
                    zip_file.writestr(self.pandataset.id + '_metadata.json', package)
                    self.logging.append({'SUCCESS': 'Frictionless in memory ZIP created'})
                except Exception as e:
                    self.logging.append({'ERROR': 'Frictionless in memory Zip creation failed: '+str(e)})
            else:
                self.logging.append({'ERROR': 'Dataset is protected'})
        else:
            self.logging.append({'ERROR':'Cannot export a parent type dataset to frictionless'})
        return in_memory_zip

    def save(self):
        if isinstance(self.file, BytesIO):
            try:
                with open(os.path.join(self.filelocation,str('frictionless_pangaea_'+str(self.pandataset.id)+'.zip')),'wb') as f:
                    f.write(self.file.getbuffer())
                    f.close()
                    return True
            except Exception as e:
                self.logging.append({'ERROR': 'Could not save Frictionless Zip: '+str(e)})
        else:
            self.logging.append({'ERROR':'Could not save, Frictionless Zip file is not a BytesIO'})
            return False


from setuptools import setup, find_packages

setup(
    name='pangaeapy',
    version='1.0.8b',
    install_requires=['lxml>=4.9.1','requests~=2.26.0','pandas~=1.3.5','numpy>=1.21.0','netcdf4~=1.5.6'],
    packages=['pangaeapy','pangaeapy.exporter'],
    package_dir={'pangaeapy': ''},
    #package_data={'mypkg': ['data/*.dat']},
    package_data={'pangaeapy': ['mappings/*.json','data/*.csv','export/*.*']},
    include_package_data=True,
    url='https://github.com/pangaea-data-publisher/pangaeapy',
    license='OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)',
    author='Robert Huber',
    author_email='rhuber@uni-bremen.de',
    description='This module allows to download and analyse metadata as well as data from tabular PANGAEA (https://www.pangaea.de) datasets.                  Usage:         import pangaeapy.pandataset as pd         ds= pd.PanDataSet(787140)         print(ds.title)         print(ds.data.head())                  Please visit the github project page to see more documentation and some examples:         https://github.com/pangaea-data-publisher/pangaeapy'
)


import pandas as pd

md = pd.read_csv('../data/parameter_mapping.csv')
md.set_index('PANGAEA ID', inplace = True)
print(md.index.dtype)
print(md.columns)
print(md.to_json('pan_mappings.json',orient='index'))